"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.PreAggregations = exports.tablesToVersionEntries = exports.getStructureVersion = exports.getLastUpdatedAtTimestamp = exports.version = exports.LAMBDA_TABLE_PREFIX = void 0;
const ramda_1 = __importDefault(require("ramda"));
const crypto_1 = __importDefault(require("crypto"));
const shared_1 = require("@cubejs-backend/shared");
const lru_cache_1 = __importDefault(require("lru-cache"));
const QueryCache_1 = require("./QueryCache");
const PreAggregationPartitionRangeLoader_1 = require("./PreAggregationPartitionRangeLoader");
const PreAggregationLoader_1 = require("./PreAggregationLoader");
const PreAggregationLoadCache_1 = require("./PreAggregationLoadCache");
/// Name of the inline table containing the lambda rows.
exports.LAMBDA_TABLE_PREFIX = 'lambda';
function encodeTimeStamp(time) {
    return Math.floor(time / 1000).toString(32);
}
function decodeTimeStamp(time) {
    return parseInt(time, 32) * 1000;
}
function version(cacheKey) {
    let result = '';
    const hashCharset = 'abcdefghijklmnopqrstuvwxyz012345';
    const digestBuffer = crypto_1.default.createHash('md5').update(JSON.stringify(cacheKey)).digest();
    let residue = 0;
    let shiftCounter = 0;
    for (let i = 0; i < 5; i++) {
        const byte = digestBuffer.readUInt8(i);
        shiftCounter += 8;
        // eslint-disable-next-line operator-assignment,no-bitwise
        residue = (byte << (shiftCounter - 8)) | residue;
        // eslint-disable-next-line no-bitwise
        while (residue >> 5) {
            result += hashCharset.charAt(residue % 32);
            shiftCounter -= 5;
            // eslint-disable-next-line operator-assignment,no-bitwise
            residue = residue >> 5;
        }
    }
    result += hashCharset.charAt(residue % 32);
    return result;
}
exports.version = version;
// Returns the oldest timestamp, if any.
function getLastUpdatedAtTimestamp(timestamps) {
    timestamps = timestamps.filter(t => t !== undefined);
    if (timestamps.length === 0) {
        return undefined;
    }
    else {
        return Math.min(...timestamps);
    }
}
exports.getLastUpdatedAtTimestamp = getLastUpdatedAtTimestamp;
function getStructureVersion(preAggregation) {
    const versionArray = [preAggregation.structureVersionLoadSql || preAggregation.loadSql];
    if (preAggregation.indexesSql && preAggregation.indexesSql.length) {
        versionArray.push(preAggregation.indexesSql);
    }
    if (preAggregation.streamOffset) {
        versionArray.push(preAggregation.streamOffset);
    }
    if (preAggregation.outputColumnTypes) {
        versionArray.push(preAggregation.outputColumnTypes);
    }
    return version(versionArray.length === 1 ? versionArray[0] : versionArray);
}
exports.getStructureVersion = getStructureVersion;
const tablesToVersionEntries = (schema, tables) => ramda_1.default.sortBy(table => -table.last_updated_at, tables.map(table => {
    const match = (table.table_name || table.TABLE_NAME).match(/(.+)_(.+)_(.+)_(.+)/);
    if (!match) {
        return null;
    }
    const entity = {
        table_name: `${schema}.${match[1]}`,
        content_version: match[2],
        structure_version: match[3],
    };
    if (match[4].length < 13) {
        entity.last_updated_at = decodeTimeStamp(match[4]);
        entity.naming_version = 2;
    }
    else {
        entity.last_updated_at = parseInt(match[4], 10);
    }
    if (table.build_range_end) {
        entity.build_range_end = table.build_range_end;
    }
    return entity;
}).filter(ramda_1.default.identity));
exports.tablesToVersionEntries = tablesToVersionEntries;
class PreAggregations {
    redisPrefix;
    driverFactory;
    logger;
    queryCache;
    options;
    externalDriverFactory;
    structureVersionPersistTime;
    touchTablePersistTime;
    dropPreAggregationsWithoutTouch;
    usedTablePersistTime;
    externalRefresh;
    loadCacheQueue = {};
    queue = {};
    getQueueEventsBus;
    touchCache;
    constructor(redisPrefix, driverFactory, logger, queryCache, options) {
        this.redisPrefix = redisPrefix;
        this.driverFactory = driverFactory;
        this.logger = logger;
        this.queryCache = queryCache;
        this.options = options || {};
        this.externalDriverFactory = options.externalDriverFactory;
        this.structureVersionPersistTime = options.structureVersionPersistTime || 60 * 60 * 24 * 30;
        this.touchTablePersistTime = options.touchTablePersistTime || (0, shared_1.getEnv)('touchPreAggregationTimeout');
        this.dropPreAggregationsWithoutTouch = options.dropPreAggregationsWithoutTouch || (0, shared_1.getEnv)('dropPreAggregationsWithoutTouch');
        this.usedTablePersistTime = options.usedTablePersistTime || (0, shared_1.getEnv)('dbQueryTimeout');
        this.externalRefresh = options.externalRefresh;
        this.getQueueEventsBus = options.getQueueEventsBus;
        this.touchCache = new lru_cache_1.default({
            max: (0, shared_1.getEnv)('touchPreAggregationCacheMaxCount'),
            maxAge: (0, shared_1.getEnv)('touchPreAggregationCacheMaxAge') * 1000,
            stale: false,
            updateAgeOnGet: false
        });
    }
    tablesUsedRedisKey(tableName) {
        // TODO add dataSource?
        return this.queryCache.getKey('SQL_PRE_AGGREGATIONS_TABLES_USED', tableName);
    }
    tablesTouchRedisKey(tableName) {
        // TODO add dataSource?
        return this.queryCache.getKey('SQL_PRE_AGGREGATIONS_TABLES_TOUCH', tableName);
    }
    refreshEndReachedKey() {
        // TODO add dataSource?
        return this.queryCache.getKey('SQL_PRE_AGGREGATIONS_REFRESH_END_REACHED', '');
    }
    async addTableUsed(tableName) {
        await this.queryCache.getCacheDriver().set(this.tablesUsedRedisKey(tableName), true, this.usedTablePersistTime);
    }
    async tablesUsed() {
        return (await this.queryCache.getCacheDriver().keysStartingWith(this.tablesUsedRedisKey('')))
            .map(k => k.replace(this.tablesUsedRedisKey(''), ''));
    }
    async updateLastTouch(tableName) {
        if (this.touchCache.has(tableName)) {
            return;
        }
        try {
            this.touchCache.set(tableName, true);
            await this.queryCache.getCacheDriver().set(this.tablesTouchRedisKey(tableName), new Date().getTime(), this.touchTablePersistTime);
        }
        catch (e) {
            this.touchCache.del(tableName);
            throw e;
        }
    }
    async tablesTouched() {
        return (await this.queryCache.getCacheDriver().keysStartingWith(this.tablesTouchRedisKey('')))
            .map(k => k.replace(this.tablesTouchRedisKey(''), ''));
    }
    async updateRefreshEndReached() {
        return this.queryCache.getCacheDriver().set(this.refreshEndReachedKey(), new Date().getTime(), this.touchTablePersistTime);
    }
    async getRefreshEndReached() {
        return this.queryCache.getCacheDriver().get(this.refreshEndReachedKey());
    }
    /**
     * Determines whether the partition table is already exists or not.
     */
    async isPartitionExist(request, external, dataSource = 'default', schema, table, key, token) {
        // fetching tables
        const loadCache = new PreAggregationLoadCache_1.PreAggregationLoadCache(() => this.driverFactory(dataSource), this.queryCache, this, {
            requestId: request,
            dataSource,
            tablePrefixes: external ? null : [schema],
        });
        let tables = await loadCache.fetchTables({
            external,
            preAggregationsSchema: schema,
        });
        tables = tables.filter(row => `${schema}.${row.table_name}` === table);
        // fetching query result
        const { queueDriver } = this.queue[dataSource];
        const conn = await queueDriver.createConnection();
        const result = await conn.getResult(key);
        queueDriver.release(conn);
        // calculating status
        let status;
        if (tables.length === 1) {
            status = 'done';
        }
        else {
            status = result && result.error
                ? `failure: ${result.error}`
                : 'missing_partition';
        }
        // updating jobs cache if needed
        if (result) {
            const preAggJob = await this
                .queryCache
                .getCacheDriver()
                .get(`PRE_AGG_JOB_${token}`);
            await this
                .queryCache
                .getCacheDriver()
                .set(`PRE_AGG_JOB_${token}`, {
                ...preAggJob,
                status,
            }, 86400);
        }
        // returning response
        return [true, status];
    }
    loadAllPreAggregationsIfNeeded(queryBody) {
        const preAggregations = queryBody.preAggregations || [];
        const loadCacheByDataSource = queryBody.preAggregationsLoadCacheByDataSource || {};
        const getLoadCacheByDataSource = (dataSource = 'default', preAggregationSchema) => {
            if (!loadCacheByDataSource[`${dataSource}_${preAggregationSchema}`]) {
                loadCacheByDataSource[`${dataSource}_${preAggregationSchema}`] =
                    new PreAggregationLoadCache_1.PreAggregationLoadCache(() => this.driverFactory(dataSource), this.queryCache, this, {
                        requestId: queryBody.requestId,
                        dataSource,
                        tablePrefixes: 
                        // Can't reuse tablePrefixes for shared refresh scheduler cache
                        !queryBody.preAggregationsLoadCacheByDataSource ?
                            preAggregations
                                .filter(p => (p.dataSource || 'default') === dataSource)
                                .map(p => p.tableName.split('.')[1]) : null
                    });
            }
            return loadCacheByDataSource[`${dataSource}_${preAggregationSchema}`];
        };
        let queryParamsReplacement = null;
        const preAggregationsTablesToTempTablesPromise = preAggregations.map((p, i) => (preAggregationsTablesToTempTables) => {
            const loader = new PreAggregationPartitionRangeLoader_1.PreAggregationPartitionRangeLoader(() => this.driverFactory(p.dataSource || 'default'), this.logger, this.queryCache, this, p, preAggregationsTablesToTempTables, getLoadCacheByDataSource(p.dataSource, p.preAggregationsSchema), {
                maxPartitions: this.options.maxPartitions,
                maxSourceRowLimit: this.options.maxSourceRowLimit,
                isJob: queryBody.isJob,
                waitForRenew: queryBody.renewQuery,
                // TODO workaround to avoid continuous waiting on building pre-aggregation dependencies
                forceBuild: i === preAggregations.length - 1 ? queryBody.forceBuildPreAggregations : false,
                requestId: queryBody.requestId,
                metadata: queryBody.metadata,
                orphanedTimeout: queryBody.orphanedTimeout,
                lambdaQuery: (queryBody.lambdaQueries ?? {})[p.preAggregationId],
                externalRefresh: this.externalRefresh
            });
            const preAggregationPromise = async () => {
                const loadResult = await loader.loadPreAggregations();
                const usedPreAggregation = {
                    ...loadResult,
                    type: p.type,
                };
                await this.addTableUsed(usedPreAggregation.targetTableName);
                if (i === preAggregations.length - 1 && queryBody.values) {
                    queryParamsReplacement = await loader.replaceQueryBuildRangeParams(queryBody.values);
                }
                return [p.tableName, usedPreAggregation];
            };
            return preAggregationPromise().then(res => preAggregationsTablesToTempTables.concat([res]));
        }).reduce((promise, fn) => promise.then(fn), Promise.resolve([]));
        return preAggregationsTablesToTempTablesPromise.then(preAggregationsTablesToTempTables => ({
            preAggregationsTablesToTempTables,
            values: queryParamsReplacement
        }));
    }
    /**
     * Determines whether range queries for the preAggregations from the
     * queryBody were cached or not.
     */
    async checkPartitionsBuildRangeCache(queryBody) {
        const preAggregations = queryBody.preAggregations || [];
        const result = await Promise.all(preAggregations.map(async (preAggregation) => {
            const { preAggregationStartEndQueries } = preAggregation;
            const invalidate = preAggregation.invalidateKeyQueries &&
                preAggregation.invalidateKeyQueries[0]
                ? preAggregation.invalidateKeyQueries[0].slice(0, 2)
                : false;
            const isCached = preAggregation.partitionGranularity
                ? (await Promise.all(preAggregationStartEndQueries.map(([query, values]) => (this.queryCache.resultFromCacheIfExists({
                    query,
                    values,
                    invalidate,
                }))))).every((res) => res?.data)
                : true;
            return {
                preAggregation,
                isCached,
            };
        }));
        return result;
    }
    async expandPartitionsInPreAggregations(queryBody) {
        const preAggregations = queryBody.preAggregations || [];
        const loadCacheByDataSource = queryBody.preAggregationsLoadCacheByDataSource || {};
        const getLoadCacheByDataSource = (dataSource = 'default', preAggregationSchema) => {
            if (!loadCacheByDataSource[`${dataSource}_${preAggregationSchema}`]) {
                loadCacheByDataSource[`${dataSource}_${preAggregationSchema}`] =
                    new PreAggregationLoadCache_1.PreAggregationLoadCache(() => this.driverFactory(dataSource), this.queryCache, this, {
                        requestId: queryBody.requestId,
                        dataSource,
                    });
            }
            return loadCacheByDataSource[`${dataSource}_${preAggregationSchema}`];
        };
        const expandedPreAggregations = await Promise.all(preAggregations.map(p => {
            const loader = new PreAggregationPartitionRangeLoader_1.PreAggregationPartitionRangeLoader(() => this.driverFactory(p.dataSource || 'default'), this.logger, this.queryCache, this, p, [], getLoadCacheByDataSource(p.dataSource, p.preAggregationsSchema), {
                maxPartitions: this.options.maxPartitions,
                maxSourceRowLimit: this.options.maxSourceRowLimit,
                waitForRenew: queryBody.renewQuery,
                requestId: queryBody.requestId,
                externalRefresh: this.externalRefresh,
                compilerCacheFn: queryBody.compilerCacheFn,
            });
            return loader.partitionPreAggregations();
        }));
        expandedPreAggregations.forEach((preAggs) => preAggs.forEach(p => {
            p.expandedPartition = true;
        }));
        return {
            ...queryBody,
            preAggregations: expandedPreAggregations.reduce((a, b) => a.concat(b), []),
            groupedPartitionPreAggregations: expandedPreAggregations
        };
    }
    async getQueue(dataSource = 'default') {
        if (!this.queue[dataSource]) {
            const queueOptions = await this.options.queueOptions(dataSource);
            if (!this.queue[dataSource]) {
                this.queue[dataSource] = QueryCache_1.QueryCache.createQueue(`SQL_PRE_AGGREGATIONS_${this.redisPrefix}_${dataSource}`, () => this.driverFactory(dataSource), (client, q) => {
                    const { preAggregation, preAggregationsTablesToTempTables, newVersionEntry, requestId, invalidationKeys, buildRangeEnd } = q;
                    const loader = new PreAggregationLoader_1.PreAggregationLoader(() => this.driverFactory(dataSource), this.logger, this.queryCache, this, preAggregation, preAggregationsTablesToTempTables, new PreAggregationLoadCache_1.PreAggregationLoadCache(() => this.driverFactory(dataSource), this.queryCache, this, {
                        requestId,
                        dataSource,
                    }), { requestId, externalRefresh: this.externalRefresh, buildRangeEnd });
                    return loader.refresh(newVersionEntry, invalidationKeys, client);
                }, {
                    concurrency: 1,
                    logger: this.logger,
                    cacheAndQueueDriver: this.options.cacheAndQueueDriver,
                    cubeStoreDriverFactory: this.options.cubeStoreDriverFactory,
                    // Centralized continueWaitTimeout that can be overridden in queueOptions
                    continueWaitTimeout: this.options.continueWaitTimeout,
                    ...queueOptions,
                    getQueueEventsBus: this.getQueueEventsBus,
                });
            }
        }
        return this.queue[dataSource];
    }
    /**
     * Returns registered queries queues hash table.
     */
    getQueues() {
        return this.queue;
    }
    getLoadCacheQueue(dataSource = 'default') {
        if (!this.loadCacheQueue[dataSource]) {
            this.loadCacheQueue[dataSource] = QueryCache_1.QueryCache.createQueue(`SQL_PRE_AGGREGATIONS_CACHE_${this.redisPrefix}_${dataSource}`, 
            // eslint-disable-next-line @typescript-eslint/no-empty-function
            () => ({}), (_, q) => {
                const { preAggregation, requestId } = q;
                const loadCache = new PreAggregationLoadCache_1.PreAggregationLoadCache(() => this.driverFactory(dataSource), this.queryCache, this, {
                    requestId,
                    dataSource,
                });
                return loadCache.fetchTables(preAggregation);
            }, {
                getQueueEventsBus: this.getQueueEventsBus,
                concurrency: 4,
                logger: this.logger,
                cacheAndQueueDriver: this.options.cacheAndQueueDriver,
                cubeStoreDriverFactory: this.options.cubeStoreDriverFactory,
                ...this.options.loadCacheQueueOptions
            });
        }
        return this.loadCacheQueue[dataSource];
    }
    static preAggregationQueryCacheKey(preAggregation) {
        return preAggregation.tableName;
    }
    static targetTableName(versionEntry) {
        if (versionEntry.naming_version === 2) {
            return `${versionEntry.table_name}_${versionEntry.content_version}_${versionEntry.structure_version}_${versionEntry.last_updated_at === '*' ? versionEntry.last_updated_at : encodeTimeStamp(versionEntry.last_updated_at)}`;
        }
        return `${versionEntry.table_name}_${versionEntry.content_version}_${versionEntry.structure_version}_${versionEntry.last_updated_at}`;
    }
    static noPreAggregationPartitionsBuiltMessage(preAggregations) {
        const expectedTableNames = preAggregations.map(p => PreAggregations.targetTableName({
            table_name: p.tableName,
            structure_version: getStructureVersion(p),
            content_version: '*',
            last_updated_at: '*',
            naming_version: 2,
        }));
        return 'No pre-aggregation partitions were built yet for the pre-aggregation serving this query and ' +
            'this API instance wasn\'t set up to build pre-aggregations. ' +
            'Please make sure your refresh worker is configured correctly, running, pre-aggregation tables are built and ' +
            'all pre-aggregation refresh settings like timezone match. ' +
            `Expected table name patterns: ${expectedTableNames.join(', ')}`;
    }
    static structureVersion(preAggregation) {
        return getStructureVersion(preAggregation);
    }
    async getVersionEntries(preAggregations, requestId) {
        const loadCacheByDataSource = {};
        const getLoadCacheByDataSource = (dataSource = 'default', preAggregationSchema) => {
            if (!loadCacheByDataSource[`${dataSource}_${preAggregationSchema}`]) {
                loadCacheByDataSource[`${dataSource}_${preAggregationSchema}`] =
                    new PreAggregationLoadCache_1.PreAggregationLoadCache(() => this.driverFactory(dataSource), this.queryCache, this, {
                        requestId,
                        dataSource,
                    });
            }
            return loadCacheByDataSource[`${dataSource}_${preAggregationSchema}`];
        };
        const firstByCacheKey = {};
        const data = await Promise.all(preAggregations.map(async (preAggregation) => {
            const { dataSource, preAggregationsSchema } = preAggregation;
            const cacheKey = getLoadCacheByDataSource(dataSource, preAggregationsSchema).tablesCachePrefixKey(preAggregation);
            if (!firstByCacheKey[cacheKey]) {
                firstByCacheKey[cacheKey] = getLoadCacheByDataSource(dataSource, preAggregationsSchema).getVersionEntries(preAggregation);
                const res = await firstByCacheKey[cacheKey];
                return res.versionEntries;
            }
            return null;
        }));
        return data.filter(res => res);
    }
    async getQueueState(dataSource) {
        const queue = await this.getQueue(dataSource);
        return queue.getQueries();
    }
    async cancelQueriesFromQueue(queryKeys, dataSource) {
        const queue = await this.getQueue(dataSource);
        return Promise.all(queryKeys.map(queryKey => queue.cancelQuery(queryKey, null)));
    }
}
exports.PreAggregations = PreAggregations;
//# sourceMappingURL=PreAggregations.js.map